# -*- coding: utf-8 -*-
import itertools
import math


def luhn(value: str) -> bool:
    """
    The Luhn algorithm or Luhn formula, also known as the "modulus 10" or "mod 10" algorithm,
    named after its creator, IBM scientist Hans Peter Luhn,
    is a simple checksum formula used to validate a variety of
    identification numbers, such as credit card numbers, IMEI numbers, National Provider Identifier numbers

    Based on: https://en.wikipedia.org/wiki/Luhn_algorithm
    """

    if not value.isdigit():
        return False

    sum_total = int(value[-1])
    length = len(value)
    parity = length % 2
    for i in range(length - 1):
        digit = int(value[i])
        if i % 2 == parity:
            digit *= 2
        if digit > 9:
            digit -= 9
        sum_total += digit

    return sum_total % 10 == 0


def fibonacci(n=None):
    """
    Returns a generator with fibonacci series.

    :param Optional[int] n: number iterations.
        Must be None or an integer: 0 <= x <= sys.maxsize
    :rtype: Generator[int]
    """

    def worker():
        a, b = 0, 1
        while True:
            yield a
            b = a + b
            yield b
            a = a + b

    it = worker()
    return itertools.islice(it, 0, n) if n is not None else it


def sieve_of_eratosthenes(n):
    """
    Primes less than n.
    Based on: https://en.wikipedia.org/wiki/Sieve_of_Eratosthenes

    :param int n: n an integer n > 1
    :rtype: Generator[int]. All prime numbers from 2 through n.
    """

    sieve = bytearray((0, 0)) + bytearray((1,)) * (n - 2)
    limit = int(math.sqrt(n) + 1)
    for p in range(2, limit):
        if sieve[p]:
            mulp = p * p
            size = len(range(mulp, n, p))
            sieve[mulp:n:p] = bytes(size)
    prime = (i for i, v in enumerate(sieve) if v)
    yield from prime



# -*- coding: utf-8 -*-
def string2bool(value, ignorecase=False):
    """
    Function to convert a string representation of
    truth to True or False.

    :param str value: value to convert.
    :param bool ignorecase: Uppercase/lowercase letters of given "value" are ignored.

    :raise: "ValueError" if "value" is anything else.
    :rtype: bool

    Examples:
        from pymince.boolean import string2bool

        string2bool("true") # --> True
        string2bool("false") # --> False

        string2bool("TRUE") # --> ValueError
        string2bool("TRUE", ignorecase=True) # --> True

        string2bool("FALSE") # --> ValueError
        string2bool("FALSE", ignorecase=True) # --> False
    """

    checking = value.lower() if ignorecase else value
    if checking == "true":
        return True
    elif checking == "false":
        return False
    else:
        raise ValueError



# -*- coding: utf-8 -*-

import functools
import warnings


def deprecated(fn):
    """
    This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used.
    http://code.activestate.com/recipes/391367-deprecated/?in=lang-python

    Examples:
        from pymince.warnings import deprecated

        @deprecated
        def check_function():
            pass

        class SomeClass:
            @deprecated
            def check_method(self):
                pass

        @deprecated
        class CheckClass:
            pass

        >> check_function() # DeprecationWarning  --> 'Deprecated "check_function".'
        >> SomeClass().check_method() #  DeprecationWarning --> 'Deprecated "check_method".'
        >> CheckClass() # DeprecationWarning  --> 'Deprecated "CheckClass".'
    """

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        warnings.warn('Deprecated "%s".' % fn.__name__, category=DeprecationWarning, stacklevel=2)

        # stacklevel=2. The printed warning will indicate and show the call site to the deprecated
        # function rather than to the location of the warnings.warn() call.
        return fn(*args, **kwargs)

    return wrapper



# -*- coding: utf-8 -*-
import csv
import functools
import itertools
import json
import operator
import zipfile

dumps = functools.partial(json.dumps, ensure_ascii=False)
dump = functools.partial(json.dump, ensure_ascii=False)
ENCODING = "utf-8"


def load_from(filename, encoding=ENCODING):
    """
    Load JSON from a file using "utf-8" encoding.

    Examples:
        from pymince.json import load_from

        dictionary = load_from("foo.json")
    """
    with open(filename, encoding=encoding) as file:
        return json.load(file)


def dump_into(filename, payload, indent=None, encoding=ENCODING):
    """
    Dump JSON to a file using "utf-8" encoding.

    Examples:
        from pymince.json import dump_into

        dump_into("foo.json", {"key": "value"})
    """
    with open(filename, "w", encoding=encoding) as file:
        dump(payload, file, indent=indent)


def dump_into_zip(zip_path, arcname, payload, indent=None):
    """
    Dump JSON into the zip archive under the name arcname.

    Examples:
        from pymince.json import dump_into_zip

        dump_into_zip("archive.zip", "foo.json", {"key": "value"})
    """
    with zipfile.ZipFile(zip_path, mode="w") as zf:
        json_string = dumps(payload, indent=indent)
        zf.writestr(arcname, json_string)


def load_from_zip(zip_path, arcname):
    """
    Load JSON from a file named "arcname" inside a zip archive.

    Examples:
        from pymince.json import load_from_zip

        dictionary = load_from_zip("archive.zip", "foo.json")
    """
    with zipfile.ZipFile(zip_path, mode="r") as zf:
        with zf.open(arcname) as file:
            return json.load(file)


def dump_from_csv(
    csv_path,
    json_path,
    /,
    *,
    fieldnames=None,
    indent=None,
    start=0,
    stop=None,
    strip=True,
    encoding=ENCODING,
):
    """
    Dump CSV file to a JSON file using "utf-8" encoding.

    :param str csv_path:
    :param str json_path:
    :param fieldnames: list of keys for the JSON
    :param str indent: JSON indent
    :param int start:
        If start is specified, will skip all preceding elements;
        otherwise, start defaults to zero.
    :param int stop:
    :param bool strip:
        Whether or not white space should be removed from the
        beginning and end of field values.
    :param str encoding: utf-8 is used by default.
    """

    def read():
        reader = csv.DictReader(csv_f, fieldnames=fieldnames)
        if strip:
            apply = operator.methodcaller("strip")
            for row in reader:
                yield {k and apply(k): (apply(v) if isinstance(v, str) else v) for k, v in row.items()}
        else:
            yield from reader

    with open(csv_path, encoding=encoding, newline="") as csv_f:
        # The official csv doc recommends opening the file with newline='' on all platforms to disable
        # universal newlines translation.

        data = itertools.islice(read(), start, stop)
        pool = tuple(data)

    dump_into(json_path, pool, indent=indent, encoding=encoding)



# -*- coding: utf-8 -*-
import xml.etree.ElementTree as etree


def iterparse(filename):
    """
    Incrementally parse XML document into ElementTree.

    This function is based on: https://github.com/python/cpython/issues/93618

    'Fix misleading hint for original ElementTree.iterparse.'
    '''
    The code below deletes a root child once it is completed, then processes and removes
    it from the memory (if nothing more references to it ofc).
    This allows to process 7GB XML with with a memory usage up to 10MB (in case of great number of root children).
    '''

    :param str filename: XML filename
    :rtype: Generator

     Examples:
        from pymince.xml import iterparse

        for event, obj in iterparse("countries.xml")
            if event == 'start'
                print(obj, obj.tag, obj.attrib, obj.text)

        >>Output<<
        <Element 'country' at 0x0000018ADF9D0CC0> country {'code': 'as', 'iso': '16'} American Samoa
        <Element 'country' at 0x0000018ADF9D0C70> country {'code': 'ad', 'iso': '20'} Andorra

    """

    parser = etree.XMLPullParser(["start", "end"])  # can be replaced with iterparse as well
    root = None
    with open(filename, encoding="utf-8") as f:
        for line in f:
            parser.feed(line)
            for event, obj in parser.read_events():
                if event == "start":
                    if root is None:
                        root = obj
                elif event == "end":
                    if len(root) > 0 and obj == root[0]:
                        del root[0]
                        # process obj

                yield event, obj

    parser.close()



# -*- coding: utf-8 -*-
import functools


class classproperty:
    """
    Decorator that converts a method with a single cls argument into a property
    that can be accessed directly from the class.

    Examples:
        from pymince.functional import classproperty

        class MyClass:
            __foo = "var"

            @classproperty
            def foo(cls):
                return cls.__foo
    """

    def __init__(self, method=None):
        self.fget = method

    def __get__(self, instance, cls=None):
        return self.fget(cls)


def pipe(*fns):
    """
    Simple pipe function implementation using function composition.

    :param fns: Functions to pipe.
    :rtype: Callable[[Any], Any]

    Examples:
        from pymince.functional import pipe

        addtwo = lambda n: n + 2
        double = lambda n: n * 2
        square = lambda n: n * n

        fn = pipe(addtwo, double, square)
        fn(1) # 36
    """

    def wrap(arg):
        return functools.reduce(lambda v, fn: fn(v), fns, arg)

    return wrap



# -*- coding: utf-8 -*-
import functools
import json
import operator
import sys

import pymince.json


def bind_json_std(encoding="utf-8"):
    """
    Decorator to call "function" passing the json read from
    "stdin" in the keyword parameter "data" and dump the json that the callback returns
    to "stdout".

    Examples:
    from pymince.std import bind_json_std

    @bind_json_std()
    def foo(data=None):
        print("Processing data from sys.stdin", data)

        result = data and {**data, "new": "value"}

        print("Result to write in sys.stdout", result)
        return result
    """

    def decorator(function):
        @functools.wraps(function)
        def apply(*args, **kwargs):
            encode_std = operator.methodcaller("reconfigure", encoding=encoding)

            encode_std(sys.stdin)
            encode_std(sys.stdout)

            data = json.load(sys.stdin)
            resp = function(*args, **kwargs, data=data)
            if resp:
                pymince.json.dump(resp, sys.stdout)

        return apply

    return decorator



# -*- coding: utf-8 -*-
import functools
import gzip
import os
import re
import shutil
import zipfile


def match_from_zip(zip_file, pattern):
    """
    Make an iterator that returns file names in the zip file that
    match the given pattern.
    Uppercase/lowercase letters are ignored.

    :param zip_file: ZipFile object or zip path.
    :param pattern: "re.Pattern" to filter filename list
    :return: Iterator with the filenames found

    Examples:
        import pymince.file
        pymince.file.match_from_zip("archive.zip", "^file") # --> file1.log file2.txt
        pymince.file.match_from_zip(zipfile.ZipFile("archive.zip"), "^file") # --> file1.log file2.txt
    """

    def match(file):
        apply = re.compile(pattern, re.IGNORECASE).match
        return iter(filter(apply, file.namelist()))

    if isinstance(zip_file, zipfile.ZipFile):
        return match(zip_file)
    elif isinstance(zip_file, str):
        with zipfile.ZipFile(zip_file) as zf:
            return match(zf)
    else:
        raise ValueError(zip_file)


def ensure_directory(path, cleaning=False):
    """
    Make sure the given file path is an existing directory.
    If it does not exist, a new directory will be created.

    :param str path:
    :param bool cleaning:
        If "cleaning" is True and a directory already exists,
        this directory and the files contained in it will be deleted.

        If "cleaning" is True and a file already exists,
        this file will be deleted.
    """
    if os.path.exists(path):
        if cleaning:
            if os.path.isdir(path):
                # Delete a directory and the files contained in it.
                shutil.rmtree(path)
            else:
                os.remove(path)  # Removes the specified file.
            os.makedirs(path)
    else:
        os.makedirs(path)


def is_empty_directory(path):
    """Check if the given path is an empty directory."""

    return os.path.exists(path) and not os.path.isfile(path) and not os.listdir(path)


def replace_extension(filename, old_ext=None, new_ext=None):
    """
    Replace filename "old_ext" with "new_ext".

    :param str filename:
    :param Optional[str] old_ext:
    :param Optional[str] new_ext:

    Examples:
        from pymince.file import replace_extension

        # remove extensions
        replace_extension("/home/user/file.old") # --> "/home/user/file"
        replace_extension("/home/user/file.old", old_ext=".old") # --> "/home/user/file"
        replace_extension("/home/user/file.old", old_ext=".new") # --> "/home/user/file.old"

        # replace extensions
        replace_extension("/home/user/file.old", new_ext=".new") # --> "/home/user/file.new"
        replace_extension("/home/user/file.old", old_ext=".old", new_ext=".new") # --> "/home/user/file.new"
        replace_extension("/home/user/file.old", old_ext=".new", new_ext=".new") # --> "/home/user/file.old"
    """

    root, ext = os.path.splitext(filename)  # Split the path in root and ext pair
    if old_ext is None or old_ext == ext:
        return root + (new_ext or "")
    else:
        return filename


def decompress(src_path, dst_path):
    """
    Decompress given file in blocks using gzip.

    :param str src_path: source file path
    :param str dst_path: destination file(unzipped) path
    :return: dst_path

     Examples:
        from pymince.file import decompress

        decompress("/foo/src.txt.gz", "/baz/dst.txt") # --> "/baz/dst.txt"
    """

    with gzip.open(src_path) as src:
        with open(dst_path, mode="wb") as dst:
            lines = iter(functools.partial(src.read, 64 * 1024), b"")
            dst.writelines(lines)
    return dst_path



# -*- coding: utf-8 -*-
"""Useful functions for working with strings."""
import functools
import html
import re
import string
import urllib.parse

import pymince.algorithm

__remove_number_commas = functools.partial(re.compile("(?<=\\d),(?=\\d{3})").sub, "")

_int_regexp_any = re.compile(r"^[-+]?[1-9]\d*\.?[0]*$")
_int_regexp_pos = re.compile(r"^[+]?[1-9]\d*\.?[0]*$")
_int_regexp_neg = re.compile(r"^-[1-9]\d*\.?[0]*$")
_bin_regexp_not = re.compile(r"[^01]")
_percentage_regexp = re.compile(r"^(?:0|[1-9]\d*)(?:\.\d+)?(?:\s)?%$")
_email_address_regexp = re.compile(r"^\S+@\S+$")
_roman_regex = re.compile(r"^M{0,3}(CM|CD|D?C{0,3})(XC|XL|L?X{0,3})(IX|IV|V?I{0,3})$")


def remove_number_commas(string):
    """
    Removes commas from a formatted text number having commas
    as group separator.

    :param str string:
    :rtype str

    Examples:
        from pymince.text import remove_number_commas
        remove_number_commas('1,234,567.8') # --> '1234567.8'
    """

    return __remove_number_commas(string)


def replace(value, old_values, new_value, count=-1):
    """
    Replace matching values ​​in the given string with new_value.

    :param str value:
    :param old_values: iterable of values ​​to replace.
    :param str new_value: replacement value.
    :param int count:
        Maximum number of occurrences to replace.
        -1 (the default value) means replace all occurrences.
    :rtype: str

    Examples:
        from pymince.text import replace

        replace("No, woman, no cry", [","], ";") # --> "No; woman; no cry"
        replace("No, woman, no cry", [","], ";", count=1) # --> "No; woman, no cry"
    """

    for old_value in old_values:
        value = value.replace(old_value, new_value, count)
    return value


def multireplace(text, replacements):
    """
    Given a string and a replacement map, it returns the replaced string.

    :param str text: string to execute replacements on
    :param Union[dict[str, any], Tuple(str, any)] replacements:
        2-dict or 2-tuple with value to find and value to replace
    :rtype: str

     Examples:
        from pymince.text import multireplace

        mapping = {",": "", "cry": "smile"}
        multireplace("No, woman, no cry", mapping) # --> "No woman no smile"

    """

    mapping = dict()
    template = text
    for i, (word, new_value) in enumerate(dict(replacements).items()):
        # set valid identifiers to template
        uid = f"id{i}"
        template = template.replace(word, f"${uid}")
        mapping[uid] = new_value
    return string.Template(template).substitute(**mapping)


def remove_decimal_zeros(value, decimal_sep=".", min_decimals=None):
    """
    Removes non-significant decimal zeros from a formatted text number.

    Examples:
        from pymince.text import remove_decimal_zeros

        remove_decimal_zeros("2.000100", ".") # --> "2.0001"
        remove_decimal_zeros("2.000000", ".") # --> "2"
        remove_decimal_zeros("2.000000", ".", min_decimals=2) # --> "2.00"
    """

    if not value:
        return ""

    int_part, _, dec_part = value.partition(decimal_sep)
    int_part = int_part or "0"  # ensure ('.1' => '0.1')
    # remove decimal zeros to the right.
    dec_part = dec_part.rstrip("0")
    dec_part_size = len(dec_part)
    # If the integer part is not 0, it is checked if it is necessary
    # to add decimals to decimal part based on "min_decimals" param
    if int_part != "0" and min_decimals and dec_part_size < min_decimals:
        dec_diff = "0" * (min_decimals - dec_part_size)
        dec_part = f"{dec_part}{dec_diff}"

    return f"{int_part}{decimal_sep}{dec_part}" if dec_part else int_part


def is_url(text, schemes=None, hostnames=None):
    """
    Check if the string is a URL according to the
    given schemes and host-names.

    :param str text:
    :param Optional[Container[str]] schemes: ("http", "https")
    :param Optional[Container[str]] hostnames: ("www.python.org", "github.com", "localhost")
    :rtype: bool

    Examples:
        from pymince.text import is_url

        # True
        is_url("https://github.com/")
        is_url("https://github.com/", hostnames=("github.com",))
        is_url("https://github.com/", hostnames=("github.com",), schemes=("https",))

        # False
        is_url("https://github.com/", schemes=("http",))
        is_url("https://github.com/", hostnames=("www.python.org", "localhost"))
    """

    res = urllib.parse.urlparse(text)
    if not res.scheme or (schemes and res.scheme not in schemes):
        return False
    if not res.netloc or (hostnames and res.hostname not in hostnames):
        return False
    try:
        _ = res.port
    except ValueError:
        return False
    else:
        return True


def is_binary(text):
    """Check if the string is binary or not."""
    return not next(_bin_regexp_not.finditer(text), None)


def is_int(text):
    """
    Check if the string is the representation of
    a integer number.

    True:
     "10",   "10.",   "10.0",
    "+10",  "+10.",  "+10.0",
    "-10",  "-10.",  "-10.0"
    """
    return _int_regexp_any.fullmatch(text) is not None


def is_positive_int(text):
    """
    Check if the string is the representation of
    positive integer number.

    True:
     "10",   "10.",   "10.0",
    "+10",  "+10.",  "+10.0",
    """
    return _int_regexp_pos.fullmatch(text) is not None


def is_negative_int(text):
    """
    Check if the string is the representation of
    negative integer number.

    True:
    "-10",  "-10.",  "-10.0"
    """

    return _int_regexp_neg.fullmatch(text) is not None


def is_payment_card(text):
    """
    Check if the string is a valid payment
    card number.

    https://en.wikipedia.org/wiki/Payment_card_number#Issuer_identification_number_(IIN)
    """

    numb = text.replace(" ", "")
    size = len(numb)
    if 12 <= size <= 19 and pymince.algorithm.luhn(numb):
        if numb[0] == "4":  # Visa
            isvalid = size in {13, 16, 19}
        elif 51 <= int(numb[:2]) <= 55:  # Mastercard
            isvalid = size == 16
        elif numb[:2] in {"34", "37"}:  # American Express
            isvalid = size == 15
        else:
            isvalid = True
        return isvalid
    else:
        return False


def is_email_address(text):
    """
    Check if the string is an email address.

    This solution does a very simple check. It only validates that the string contains an at sign (@)
    that is preceded and followed by one or more non whitespace characters.
    """
    return _email_address_regexp.fullmatch(text) is not None


def is_percentage(text):
    """
    Check if the string is a valid percentage

    True: "100%", "100 %", "100&nbsp;%", 100.0 %",
    """
    unescaped = html.unescape(text)  # 100&nbsp;% => 100 %
    return _percentage_regexp.fullmatch(unescaped) is not None


def is_palindrome(text):
    """
    Check if the string is palindrome or not.
    A string is said to be palindrome if the reverse of the string is the same as string
    """
    return text == text[::-1]


def is_roman(text):
    """Check if the string is a valid roman numeral."""
    return _roman_regex.fullmatch(text) is not None


class fullstr(str):
    """
    Custom string inheriting from "str" which adds
    the following methods:

    - is_url(self, schemes=None, hostnames=None)
    - is_int(self)
    - is_positive_int(self)
    - is_negative_int(self)
    - is_payment_card(self)
    - is_binary(self)
    - is_percentage(self)
    - is_palindrome(self)
    - is_email_address(self)
    - is_roman(self)
    """

    def is_url(self, schemes=None, hostnames=None):
        return is_url(self, schemes=schemes, hostnames=hostnames)

    def is_binary(self):
        return is_binary(self)

    def is_int(self):
        return is_int(self)

    def is_positive_int(self):
        return is_positive_int(self)

    def is_negative_int(self):
        return is_negative_int(self)

    def is_payment_card(self):
        return is_payment_card(self)

    def is_email_address(self):
        return is_email_address(self)

    def is_percentage(self):
        return is_percentage(self)

    def is_palindrome(self):
        return is_palindrome(self)

    def is_roman(self):
        return is_roman(self)



# -*- coding: utf-8 -*-
from setuptools import setup

from pymince import __version__, __title__


def read(filename):
    with open(filename, encoding="utf-8") as f:
        return f.read()


setup(
    name=__title__,
    version=__version__,
    description="Python shredded utilities",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Internet",
        "Topic :: Utilities",
    ],
    keywords="",
    author="rmoralespp",
    author_email="rmoralespp@gmail.com",
    url="https://github.com/rmoralespp/pymince",
    license="MIT",
    packages=["pymince"],
    include_package_data=True,
    zip_safe=False,  # https://mypy.readthedocs.io/en/latest/installed_packages.html
    # install_requires=read('requirements.txt'),
    python_requires=">=3.8",
)



# -*- coding: utf-8 -*-
import contextlib
import datetime


def irange(start_date, stop_date=None, time_step=None):
    """
    Returns a generator that produces a sequence of datetime's from "start_date" (inclusive)
    to "stop_date" (exclusive) by "time_step".

    :param datetime.datetime start_date: Inclusive.
    :param datetime.datetime stop_date: Exclusive. `utcnow` is used by default.
    :param datetime.delta time_step: one-day `timedelta` is used by default.

     Examples:
        import datetime

        from pymince.dates import irange

        ini = datetime.datetime.fromisoformat("2022-10-31")
        end = datetime.datetime.fromisoformat("2022-11-02")
        day = datetime.timedelta(days=1)

        it = irange(ini, stop_date=end, time_step=day)

        next(it) # --> datetime.datetime(2022, 10, 31, 0, 0)
        next(it) # --> datetime.datetime(2022, 11, 1, 0, 0)
        next(it) # --> raise StopIteration
    """

    this = start_date
    stop = stop_date or datetime.datetime.utcnow()
    step = time_step or datetime.timedelta(days=1)
    while this < stop:
        yield this
        this += step


def string2year(value, gte=None, lte=None, shift=None):
    """
    Function to convert a string year representation to integer year.

    :param str value: Value to convert.
    :param Optional[int] gte: if it is specified is required that: year >= gte
    :param Optional[int] lte: if it is specified is required that: year <= lte
    :param Optional[int] shift: use a two-digit year on shift

    :raise: "ValueError" if "value" cannot be converted.
    :rtype: int

    Examples:
        from pymince.dates import string2year

        string2year("53", shift=None) # --> 2053
        string2year("53", shift=1953) # --> 1953
        string2year("52", shift=1953) # --> 2052
        string2year("54", shift=1953) # --> 1954

        string2year("1954") # --> 1954

        string2year("123") # --> ValueError
        string2year("1955", gte=1956) # --> ValueError
        string2year("1955", lte=1954) # --> ValueError
    """

    year = None
    for rep in ("%Y", "%y"):
        with contextlib.suppress(ValueError, TypeError):
            year = datetime.datetime.strptime(value, rep).year
            if shift and rep == "%y":
                # Use shift for year of two-digit format
                year = (year - shift) % 100 + shift
            if (gte and not year >= gte) or (lte and not year <= lte):
                year = None
            break

    if year:
        return year
    else:
        raise ValueError



# -*- coding: utf-8 -*-
"""Useful functions that use dictionaries."""

import datetime
import enum
import functools
import hashlib
import json
import operator
import types


def all_true_values(dictionary, keys):
    """
    Check if an dictionary has all specified keys and
    key-related values as True.

    :param dict dictionary:
    :param keys: keys sequence
    :rtype: bool

    Examples:
    from pymince.dictionary import all_true_values

    all_true_values({"a": 1, "b": 2}, ("a", "b")) # --> True
    all_true_values({"a": 1, "b": 0}, ("a", "b")) # --> False
    all_true_values({"a": 1, "b": 0}, ("a",)) # --> True
    """

    getter = operator.itemgetter(*keys)
    try:
        values = getter(dictionary)
    except KeyError:
        return False
    else:
        return all(values) if len(keys) > 1 else bool(values)


def find_leaf_value(key, dictionary):
    """
    Find leaf value in mapping.

    :param str key:
    :param dict dictionary:

    Examples:
        from pymince.dictionary import find_leaf_value

        find_leaf_value('a', {}) # --> 'a'
        find_leaf_value('a', {'a': 'b', 'b': 'c'}) # --> 'c'
        find_leaf_value('a', {'a': 'a'}) # --> 'a'
    """
    while True:
        if key is not None and key in dictionary:
            new_key = dictionary[key]
            if new_key == key:
                break
            key = new_key
        else:
            break
    return key


class DigestGetter:
    """
    Calculate a digest of a "jsonified" python dictionary.

    :param include_keys: dictionary keys to exclude
    :param exclude_keys: dictionary keys to include
    :rtype: str

    Examples:
        from pymince.dictionary import DigestGetter

        getter = DigestGetter(include_keys=("a",))
        getter({"a": 1, "b": 1}) # --> bb6cb5c68df4652941caf652a366f2d8
        getter({"a": 1}) # --> bb6cb5c68df4652941caf652a366f2d8
    """

    def __init__(self, include_keys=None, exclude_keys=None):
        if include_keys and exclude_keys:
            raise ValueError
        self.include = include_keys and frozenset(include_keys)
        self.exclude = exclude_keys and frozenset(exclude_keys)

    def __call__(self, dictionary):
        string = self.to_string(dictionary)
        return hashlib.md5(string.encode("utf-8")).hexdigest()

    @functools.cached_property
    def stringify(self):
        """
        Return a function to encode a dict into json using the most compact form.
        Dictionary keys are sorted.
        Encodes datetime objects into isoformat strings.
        Encodes enum objects according to "value" attribute.
        Encode sets by ordering their values.
        """

        class Encoder(json.JSONEncoder):
            def default(self, obj):
                if isinstance(obj, datetime.datetime):
                    return obj.isoformat()
                elif isinstance(obj, enum.Enum):
                    return obj.value
                elif isinstance(obj, (frozenset, set)):
                    return sorted(obj)
                else:
                    return super().default(obj)

        return Encoder(
            separators=(",", ":"),
            check_circular=False,
            sort_keys=True,
            ensure_ascii=False,
        ).encode

    def to_string(self, dictionary):
        # Non recursive copy.
        if self.include:
            data = {k: v for k, v in dictionary.items() if k in self.include}
        elif self.exclude:
            data = {k: v for k, v in dictionary.items() if k not in self.exclude}
        else:
            data = dictionary
        return self.stringify(data)


def frozendict(*args, **kwargs):
    """
    Returns a "MappingProxyType" from a dictionary built according to given parameters.
    Add immutability only on a first level.

    Examples:
        from pymince.dictionary import frozendict

        my_dict = frozendict(a=1, b=2)
        my_dict["a"] # --> 1
        list(my_dict.items())  # --> [("a", 1), ("b", 2)]
        my_dict["c"] = 3  # --> TypeError
    """
    return types.MappingProxyType(dict(*args, **kwargs))


def from_objects(iterable, key_getter, value_getter):
    """
    Create a new dictionary with elements generated from
    the "key_getter" and "value_getter" callbacks applied to each element of the iterable.

    :param Iterable[any] iterable:
    :param Callable key_getter:
        Dictionary keys getter.
        It is called with each element of "iterable" passing it as an argument.
    :param Callable value_getter:
        Dictionary values getter.
        It is called with each element of "iterable" passing it as an argument.

    :raise: ValueError if any generated key is duplicate.
    :rtype: dict

    Examples:
        from pymince.dictionary import from_objects

        keygetter = operator.itemgetter(0)
        valgetter = operator.itemgetter(1, 2)

        values = iter([(1, "a", "b"), (2, "a", "b")])
        from_objects(values, keygetter, valgetter) # --> {1: ('a', 'b'), 2: ('a', 'b')}
    """

    dictionary = {}
    for obj in iter(iterable):
        key = key_getter(obj)
        if key in dictionary:
            raise ValueError
        else:
            dictionary[key] = value_getter(obj)
    return dictionary
